#include "Cstring.h"

// Default constructor - constructs an empty string of MAX_LENGTH.
// Postcond: An empty string of MAX_LENGTH has been constructed, maxSize is set as 255.
Cstring::Cstring()
{	str = new char[MAX_LENGTH+1];
	str[0]='\0';
	maxSize = MAX_LENGTH;
}

// Constructor - constructs an empty string of the specified length,len.
// Precond:  length must be greater than zero.
// Postcond: maxSize is length, and an empty string of the specified length has been 
//			 constructed.
Cstring::Cstring(int len)
{	str = new char[len+1];
	str[0]='\0';
	maxSize = len;
}

// Copy constructor - constructs a string that has the same maxSize and value as fromStr
// Postcond: self is a copy (both value and maxSize) of fromStr.
Cstring::Cstring(const Cstring & fromStr)
{	int L;
	L = strlen(fromStr.str);
	str = new char[L+1];
	strcpy(str, fromStr.str);
	maxSize = L;
}

// Converts a C-style string to a Cstring.
// Precond:  fromStr must be a null terminated character array (i.e. a C style string)
// Postcond: The value part of self is fromStr and maxSize is the length of fromStr
Cstring::Cstring(const char *fromStr)
{	int L;
	L = strlen(fromStr);
	str = new char[L+1];
	strcpy(str, fromStr);
	maxSize = L;
}

// Destructor releases dynamic storage for str
//Cstring::~Cstring()
//{	delete [] str;
//}

// Converts a Cstring object to a C style string, by returning the eqivalent of str
char * Cstring::c_str()const
{	char * s;
	s = new char[maxSize+1];
	strcpy (s, str);
	return s;
}

// Determines the length of str.
int Cstring::length()const
{	return strlen(str);
}

// Looks for first occurrence of c in str, beginning at index start. Return the 
// index of c in str if match occurs; otherwise return -1.
// Precond: start if specified must be greater than or equal to 0
int Cstring::findFirstOf(char c, int start)const
{	unsigned int i;
	char l;

	for (i=start; i<strlen(str); i++)
	{	l = str[i];
		if (l==c)
			return i;
	}
	return -1;
}

// Looks for the last occurrence of c in str. Return the index of the match if it 
// occurs; otherwise return -1.
int Cstring::findLastOf(char c)const
{	unsigned int i;
	char l;
	for (i=strlen(str)-1; i>=0; i--)
	{	l = str[i];
		if (l==c)
			return i;
	}
	return -1;
}

// Copy count characters from self beginning at index start, and return the resulting 
// string as a Cstring object. If the tail of str has fewer than count characters or 
// count is -1 the rest of the string is copied stopping at the end of the string.
// Precond: start must be greater than or equal to 0. count must be greater than or 
//			equal to -1
Cstring Cstring::substr(int start, int count)const
{	int thisL;
	thisL = strlen(str);
	
	// determine number of characters in rest of string
	if (count == -1)				
		count = thisL-start;
	// if count runs over end of string, reset count to end of string
	if (count > thisL-start)		
		count = thisL-start;

	Cstring temp(count);
	for (int i=0; i<=count; i++)
		temp.str[i] = str[start+i];
	temp.str[count] = '\0';						// make C-style string
	temp.maxSize = count;

	return temp;
}

// Locates a specific pattern in the string.  The method takes a Cstring object s and 
// and a index start as arguments and looks for mathc of s as a substring in self. It
// returns the starting index of the match if it occurs; otherwise it returns -1.
int Cstring::find(const Cstring &s, int start)const
{	bool keepGoing;
	int thisL, sL;
	thisL = strlen(str);
	sL = strlen(s.str);

	// begin comparison at start index
	for (int i=start; i < thisL-sL ; i++)
	{	if (str[i] == s.str[0])						// compare with first letter of s
			keepGoing = true;
		else 
			keepGoing = false;
		for (int j=1; j<sL && keepGoing; j++)		// continue comparison as long as it matches
			if (str[i+j] != s.str[j])
				keepGoing = false;
		if (keepGoing)								// match found
			return i;
	}
	return -1;										// match not found
}

// Locates a specific pattern in the string.  The method takes a C-style string  s and 
// and a index start as arguments and looks for mathc of s as a substring in self. It
// returns the starting index of the match if it occurs; otherwise it returns -1.
int Cstring::find(const char *s, int start)const
{	bool keepGoing;
	int thisL, sL;
	thisL = strlen(str);
	sL = strlen(s);
	for (int i=start; i < thisL-sL ; i++)
	{	if (str[i] == s[0])							// compare with first letter of s
			keepGoing = true;
		else 
			keepGoing = false;
		for (int j=1; j<sL && keepGoing; j++)		// continue comparison as long as it matches
			if (str[i+j] != s[j])
				keepGoing = false;
		if (keepGoing)								// match found
			return i;
	}
	return -1;										// match not found
}

// Place the substring s into the string beginning at index start.  The insertion expands
// the maxSize of the string if necessary.
// Precond: start must be greater than or equal to 0 and less than or equal to the length 
//			of str.
// Postcond: The substring s has been placed into str at the specified position. The 
//			 maxSize of self has been increased if necessary.
void Cstring::insert(int start, const Cstring &s)
{	int thisL, sL, newL, i;
	thisL  = strlen (str);

	sL = strlen(s.str);

	newL = thisL + sL;
	char * temp;						

	// determine size of temp
	if (newL > maxSize)
		temp = new char[newL+1];			// create temp C-style string
	else 
		temp = new char[maxSize];			// create temp C-style string
	
	strcpy(temp, str);						// copy all of str into temp

	for (i=0; i<sL; i++)					// copy s.str into temp at start index
		temp[i+start] = s.str[i];

	for (i=start; i<=thisL; i++)			// copy tail of str into temp after s, including '\0'
		temp[sL+i] = str[i];
	
	if (newL > maxSize)						// if str is too small make it larger
	{	delete [] str;
		str = new char[newL+1];
		maxSize = newL;
	}
	
	strcpy(str, temp);						// copy temp into str
	delete temp;							// free storage allocated above

}

// Delete count characters from str, begining at index start. If fewer than count 
// characters exist in the tail of the string or if count is -1, delete characters 
// up to the end of the string.
// Precond: start must be greater than or equal to 0
// Postcond: The specified number of characters has been removed from str.
void Cstring::erase(int start, int count)
{	int tail;
	tail = strlen(str) - start;
	// erase tail of string
	if ((count == -1) || (count > tail))
		str[start] = '\0';
	// move characters after erasure up including '\0' character
	else
		for (int i = 0; i < tail-count; i++)
			str[start+i] = str[start+i+count];
}


// Overloads operation +. Concatenates string s to self by copying s after self. 
// s is left unchanged. maxSize of self is increased if necessary.
Cstring & Cstring::operator +(const Cstring & s)
{	int thisL, sL, newL;
	thisL = strlen(str);
	sL = strlen(s.str);
	newL = thisL + sL;
	
	// if str is too small, make larger preparatory to concatenation
	if (newL > maxSize)
	{	char * temp;						
		temp = new char[newL+1];			// create temp C-style string
		strcpy(temp,str);					// copy str into temp
		delete [] str;							// destroy str
		str = new char[newL+1];				// allocate larger C-style string for str
		maxSize = newL;						// update maxSize
		strcpy(str, temp);					// copy temp into str
		delete temp;						// free storage allocated above
	}

	// perform concatenation	
	for (int i = 0; i <= sL; i++)
		str[i+thisL] = s.str[i];

	return *this;
}

// Overloads operation +. Concatenates C-style string s to self by copying s after self. 
// s is left unchanged. maxSize of self is increased if necessary.
Cstring & Cstring::operator+(const char * s)
{	int thisL, sL, newL;
	thisL = strlen(str);
	sL = strlen(s);
	newL = thisL + sL;
	
	// if str is too small, make larger preparatory to concatenation
	if (newL > maxSize)
	{	char * temp;						
		temp = new char[newL+1];			// create temp C-style string
		strcpy(temp,str);					// copy str into temp
		delete [] str;							// destroy str
		str = new char[newL+1];				// allocate larger C-style string for str
		maxSize = newL;						// update maxSize
		strcpy(str, temp);					// copy temp into str
		delete temp;						// free storage allocated above
	}

	// perform concatenation	
	for (int i = 0; i <= sL; i++)
		str[i+thisL] = s[i];

	return *this;
}


// Overloads operation +. Concatenates character c to self by copying c after self. 
// c is left unchanged. maxSize of self is increased if necessary.
Cstring & Cstring::operator+(const char c)
{	int thisL, newL;
	thisL = strlen(str);
	newL = thisL + 1;
	
	// if str is too small, make larger preparatory to concatenation
	if (newL > maxSize)
	{	char * temp;						
		temp = new char[newL+1];			// create temp C-style string
		strcpy(temp,str);					// copy str into temp
		delete [] str;							// destroy str
		str = new char[newL+1];				// allocate larger C-style string for str
		maxSize = newL;						// update maxSize
		strcpy(str, temp);					// copy temp into str
		delete temp;						// free storage allocated above
	}

	// perform concatenation	
	str[thisL] = c;
	str[thisL+1] = '\0';

	return *this;
}



// Overloads operation +=. Concatenates string s to self by copying s after self. 
// s is left unchanged. maxSize of self is increased if necessary.
Cstring & Cstring::operator +=(const Cstring & s)
{	int thisL, sL, newL;
	thisL = strlen(str);
	sL = strlen(s.str);
	newL = thisL + sL;
	
	// if str is too small, make larger preparatory to concatenation
	if (newL > maxSize)
	{	char * temp;						
		temp = new char[newL+1];			// create temp C-style string
		strcpy(temp,str);					// copy str into temp
		delete [] str;							// destroy str
		str = new char[newL+1];				// allocate larger C-style string for str
		maxSize = newL;						// update maxSize
		strcpy(str, temp);					// copy temp into str
		delete temp;						// free storage allocated above
	}
	// perform concatenation	
	for (int i = 0; i <= sL; i++)
		str[i+thisL] = s.str[i];

	return *this;
}

// Overloads operation <.Returns true if self is less than the parameter.
bool Cstring::operator < (const Cstring & s)const
{	if (strcmp(str, s.str) < 0)
		return true;
	else return false;
}

// Overloads operation <=. Returns true if self is less than or equal to the parameter.
bool Cstring::operator <= (const Cstring & s)const
{	if (strcmp(str, s.str) <= 0)
		return true;
	else return false;
}

// Overloads operation >. Returns true if self is greater than the parameter.
bool Cstring::operator > (const Cstring & s)const
{	if (strcmp(str, s.str) > 0)
		return true;
	else return false;
}

// Overloads operation >=. Returns true if self is greater than or equal to the parameter.
bool Cstring::operator >= (const Cstring & s)const
{	if (strcmp(str, s.str) >= 0)
		return true;
	else return false;
}


// Overloads operation ==. Returns true if self is equivalent the parameter.
bool Cstring::operator == (const Cstring & s)const
{	if (strcmp(str, s.str) == 0)
		return true;
	else return false;
}

// Overloads operation !=. Returns true if self is not equivalent to the parameter.
bool Cstring::operator != (const Cstring & s)const
{	if (strcmp(str, s.str) != 0)
		return true;
	else return false;
}

// Overloads the assignment operator, returning a pointer to the object self.
// Changing the size of the character array if necessary
Cstring & Cstring::operator = (const Cstring &s)
{	if  (s.maxSize != maxSize)
	{	delete [] str;
		str = new char[s.maxSize+1];
		maxSize = s.maxSize;
	}
	strcpy(str, s.str);
	return *this;
}

// Overloads operation []. Returns the character specified by the index of str
char &Cstring::operator [] (int i)
{	int thisL = strlen(str);
	if ((i < thisL) && (i >= 0))
		return str[i];
	else 
		return str[thisL];
}


// friend functions to overload I/O operators and methods
istream & operator >> (istream& istr, Cstring &s)
{	istr >> s.str;
	return istr;
}

// getline method to read an entire line until '\n'
// Precond: maxLen must be less than or equal to maxSize, if maxLen is -1 then use maxSize
// Postcond:  self contains the line which has been input.
istream & getline(istream &istr, Cstring &s)
{
	istr.getline(s.str, s.maxSize, '\n');
	return istr;
}

ostream & operator << (ostream& ostr, Cstring &s)
{	ostr << s.str;
	return ostr;
}

