//d_vector.h
//CSE124C Section 4 Lab4 by Haibei Zhang

//This class is provided by the textbook, I modified 4 methods as follows:
//T& back()   ->   T back()
//const T& back() const   ->   const T back() const
//T& operator[] (int i)    ->    T operator[] (int i)
//const T& operator[] (int i) const   ->   const T operator[] (int i) const
//Otherwise I get a compile time error when handling an int vector:"cannot conver from int to int&"

#ifndef VECTOR
#define VECTOR

#include <iostream.h>

template <typename T>
class Vector
{
	public:
		Vector(int size = 0, T item = T());
			// constructor.
			// Postconditions: allocates array with size number of elements
			// and capacity. elements are initialized to T(), the default
			// value for type T

		Vector(const Vector<T>& obj);
			// copy constructor
			// Postcondition: creates current vector as a copy of obj

		~Vector();
			// destructor
			// Postcondition: the dynamic array is destroyed

		Vector& operator= (const Vector<T>& rhs);
			// assignment operator.
			// Postcondition: current vector holds the same data
			// as rhs

		T back();
			// return the element at the rear of the vector.
			// Precondition: the vector is not empty. if vector
			// is empty, prints underflowError message and returns T()

		const T back() const;
			// const version used when Vector object is a constant

		T operator[] (int i);
			// provides general access to elements using an index.
			// Precondition: 0 <= i < vSize. if the index is out
			// of range, prints indexRangeError and returns T()

		const T operator[] (int i) const;
			// const version used when Vector object is a constant

		void push_back(const T& item);
			// insert item at the rear of the vector.
			// Postcondition: the vector size is increased by 1

		void pop_back();
			// remove element at the rear of the vector.
			// Precondition: vector is not empty. if the vector is
			// empty, prints  underflowError message and does not 
			// modify the vector

		int size() const;
			// return current list size

		bool empty() const;
			// return true if vector is empty and false otherwise

		int capacity() const;
			// return the current capacity of the vector

		bool resize(int newsize);
			// changes the size of the vector the old vector is destroyed

   private:
		int vCapacity;		// amount of available space
		int vSize;			// number of elements in the list
		T *vArr;			// the dynamic array

		bool reserve(int n, bool copy);
			// called by public functions only if n > vCapacity. expands
			// the vector capacity to n elements, copies the existing
			// elements to the new space if copy == true, and deletes
			// the old dynamic array returns true. Prints 
			// memoryAllocationError and returns false if memory 
			// allocation fails
};

// set the capacity to n elements
template <typename T>
bool Vector<T>::reserve(int n, bool copy)
{
	T *newArr;
	int i;

	// allocate a new dynamic array with n elements
	newArr = new T[n];
	if (newArr == NULL)
	{   cerr << "miniVector reserve(): memory allocation failure";
		return false;
	}
	// if copy is true, copy elements from the old list to the new list
	if (copy)
		for(i = 0; i < vSize; i++)
			newArr[i] = vArr[i];

	// delete original dynamic array. if vArr is NULL, the vector was
	// originally empty and there is no memory to delete
	if (vArr != NULL)
		delete [] vArr;

	// set vArr to the value newArr. update vCapacity
	vArr = newArr;
	vCapacity = n;
	return true;
}

// constructor. initialize vSize and vCapacity.
// allocate a dynamic array of vSize integers
// and initialize the array with T()
template <typename T>
Vector<T>::Vector(int size, T item):
	vSize(0), vCapacity(0), vArr(NULL)
{
	int i;
	bool ok=true;

	// if size is 0, vSize/vCapacity are 0 and vArr is NULL.
	// just return
	if (size == 0)
		return;

	// set capacity to size. since we are building the vector,
	// copy is false
	if (ok = reserve(size, false))
		// assign size to vSize
		vSize = size;

	if (ok)
		// copy T() into each vector element
		for (i=0;i < vSize;i++)
			vArr[i] = item;
}

// copy constructor. make the current object a copy of obj.
// for starters, use initialization list to create an empty
// vector
template <typename T>
Vector<T>::Vector (const Vector<T>& obj):
	vSize(0), vCapacity(0), vArr(NULL)
{
    int i;
	bool ok;
	// if size is 0, vSize/vCapacity are 0 and vArr is NULL.
	// just return
	if (obj.vSize == 0)
		return;

	// set capacity to obj.vSize. since we are building the vector,
	// copy is false
	if (ok=reserve(obj.vSize, false))
		// assign size to obj.vSize
		vSize = obj.vSize;
	
	if (ok)
		// copy items from the obj.vArr to the newly allocated array
		for (i = 0; i < vSize; i++)
			vArr[i] = obj.vArr[i];
}

// destructor. deallocate the dynamic array
template <typename T>
Vector<T>::~Vector()
{
	if (vArr != NULL)
		// de-allocate memory for the array
		delete [] vArr;
}

// replace existing object (left-hand operand) by
// rhs (right-hand operand)
template <typename T>
Vector<T>& Vector<T>::operator= (const Vector<T>& rhs)
{
    int i;
	bool ok=true;

    // check vCapacity to see if a new array must be allocated
    if (vCapacity < rhs.vSize)
		// make capacity of current object the size of rhs. don't
		// do a copy, since we will replace the old values
		if (ok=reserve(rhs.vSize, false))

			// assign current object to have same size as rhs
			vSize = rhs.vSize;

	if (ok)
		// copy items from rhs.vArr to vArr
		for (i = 0; i < vSize; i++)
			vArr[i] = rhs.vArr[i];

	return *this;

	
}

// check vSize and print underflowError error message if the
// value is 0 returning T(); otherwise return the element vArr[vSize-1]
template <typename T>
T Vector<T>::back()
{	
	if (vSize == 0)
	{	cerr << "Vector back(): vector empty";
		return  T();
	}

	return vArr[vSize-1];
}

template <typename T>
const T Vector<T>::back() const
{
	if (vSize == 0)
	{	cerr << "Vector back(): vector empty";
		return const T();
	}

	return vArr[vSize-1];
}

// provides general access to array elements
template <typename T>
T Vector<T>::operator[] (int i)
{
	if (i < 0 || i >= vSize)
	{	cerr << "Vector: index range error", i, vSize;
		return T();
	}

	return vArr[i];
}

// provides general access to array elements. constant version
template <typename T>
const T Vector<T>::operator[] (int i) const
{
	if (i < 0 || i >= vSize)
		throw indexRangeError(
			"miniVector: index range error", i, vSize);

	return vArr[i];
}

// insure that list has sufficient capacity,
// add the new item to the list, and increment vSize
template <typename T>
void Vector<T>::push_back(const T& item)
{
	bool ok=true;
	// if space is full, allocate more capacity
	if (vSize == vCapacity)
	{
		if (vCapacity == 0)
			// if capacity is 0, set capacity to 1.
			// set copy to false because there are
			// no existing elements
			ok = reserve(1,false);
		else
			// double the capacity
			ok = reserve(2 * vCapacity, true);
	}

	if (ok)
	{	// add item to the list, update vSize
		vArr[vSize] = item;
		vSize++;
	}
}

// if not empty, just decrement the size
template <typename T>
void Vector<T>::pop_back()
{
	if (vSize == 0)
		cerr << "miniVector pop_back(): vector is empty";
	
	else vSize--;
}

template <typename T>
int Vector<T>::size() const
{
	return vSize;
}

template <typename T>
bool Vector<T>::empty() const
{
	return vSize == 0;
}

template <typename T>
int Vector<T>:: capacity() const
{
	return vCapacity;
}

template <typename T>
bool resize(int newsize)
{
	delete [] varr;
	vCapacity = newsize;
	vSize = newsize;
	varr = new T[newsize];
	if (varr) return true;
	else return false;
}

#endif   // VECTOR
